//---------------------------------------------------------------------------

#ifndef spriteH
#define spriteH

#include <vector>
#include <typeinfo>

#include "util.h"

#undef VAR
#undef INI

#ifdef spriteCPP
#define VAR
#define INI(x) = x
#else
#define VAR extern
#define INI(x)
#endif

//---------------------------------------------------------------------------
enum TDirection {drUp, drDown, drLeft, drRight}; // V

// ݩ
enum TSpriteAttrElement {
    saUndirectionalBitmap, // ϧΤV
    saNoCellCollision, // ϮI
    saNoTankCollision, // ZJI
    saNoBulletCollision, // luI
    saAlignWithTerrItem} ; // ʮɷ|۰ʹaΪ, _Ӥ

typedef Set<TSpriteAttrElement, saUndirectionalBitmap, saAlignWithTerrItem> TSpriteAttr;

// ʵeO_έY}l
enum TAdvanceFrameMode {afWrap, afStop} ;

// _
enum TGemKind {gkClock, gkHat, gkArrow, gkStar, gkBlow, gkApple} ;

// forward declaration
class TSprite;
class TTank;
class TBullet;
class TGem;

// class declared in tileunit unit
class TCell;

// Ǧ^IG}CA
typedef std::vector<TCell*> TCellArray;

typedef std::vector<TSprite*> TSpriteArray;

// beW樫ʪ⪫
class TSprite {
private:
  int FX, FY; // y
  int FFrameNo; // ثeܪ frame s
  int FCollisionCount; // sI

  bool FActive; // O_iʧ@
  bool FVisible; // O_i
  bool FOnAir; // O_bѪ
  int FSpeed; // it

  TRect FRect; // ίxΰϰ
  TDirection FDirection; // iV
  TSpriteAttr FAttr; // ݩ

  bool FPostToDead; // O_wnOnR

  Graphics::TBitmap *FBits, *FInvBitmap; // ϤζKϥιϤ

  // ⤤IҦbϮm
  int GetTile_X();
  int GetTile_Y();

  // o⪺eΰ
  int GetObjectWidth();
  int GetObjectHeight();

  void SetDirection(TDirection Value);

  // helper functions for CheckCollisions()
  bool IsRealCellCollision(int Layer, int CellXPos, int CellYPos, int& x, int& y);
  void CheckCellCollisionsPos(TCellArray& Collisions, const TRect& SpriteRect, int Layer, int CellXPos, int CellYPos, int& x, int& y);
protected:
  // P⦳PT
  TRect FObjectRect; // ؤo
  AnsiString FFileName; // ϧɦW
  int FFrameMax; // ʵeؼ

  int FMoveDelay, FMoveDelayCount; // Uʧ@e𦸼
  TAdvanceFrameMode FAdvanceFrameMode; // ʵeBz覡

  virtual void ResetStatus(); // ]⪬A

  // IˬdĲok, tdIsҦIˬdk
  virtual bool CheckCollisions(int& x, int& y);

  // ɸIˬd
  virtual bool CheckBoundCollisions(int& x, int& y);
  // aΪIˬd
  virtual bool CheckCellCollisions(int Layer, int& x, int& y, TCellArray& Collisions);
  // Iˬd
  virtual bool CheckSpriteCollisions(TSpriteArray& Sprites, int& x, int& y, TSpriteArray& Collisions);
public:
  TSprite();
  virtual ~TSprite();

  void LoadBits(); // Jϧ

  virtual void Draw(TCanvas* Canvas); // øs

  virtual void Move(); // iU@Bʧ@

  void PostToDie(WPARAM Param = 0); // nOR

  void RandomDirection(); // HNܤV
  void RandomPosition(); // HN\m
  void CenterWith(TSprite& ASprite); // Pt@m
  void CenterBy(int x, int y); // ϤI (X, Y)

  __property int PosX = {read = FX, write = FX};
  __property int PosY = {read = FY, write = FY};

  __property int Tile_X = {read = GetTile_X};
  __property int Tile_Y = {read = GetTile_Y};

  __property TDirection Direction = {read = FDirection, write = SetDirection};
  __property int Speed = {read = FSpeed, write = FSpeed};
  __property bool Active = {read = FActive, write = FActive};
  __property bool Visible = {read = FVisible, write = FVisible};
  __property bool OnAir = {read = FOnAir, write = FOnAir};
  __property bool PostToDead = {read = FPostToDead};
  __property int CollisionCount = {read = FCollisionCount};

  __property TRect Rect = {read = FRect};
  __property TSpriteAttr Attr = {read = FAttr, write = FAttr};

  __property TRect ObjectRect = {read = FObjectRect};
  __property int ObjectWidth = {read = GetObjectWidth};
  __property int ObjectHeight = {read = GetObjectHeight};
};

// ZJO, ڤμĤZJұqO~
class TTank : public TSprite {
private:
  int FHP; // ˥ (ͩRO)
  int FBulletNum; // ثelu
  bool FSuperMode; // LļҦ

  friend class TBullet;
protected:
  // Cب⤣P
  int FScore; // QR, Do
  int FMaxBulletNum; //   P@ɶluƤW
  int FBulletBlowRange; // luz¤O (d)

  // IˬdĲok, tdIsҦIˬdk
  virtual bool CheckCollisions(int& x, int& y);
  
  // sWZJIˬd
  virtual bool CheckTankCollisions(int& x, int& y, TSpriteArray& Collisions);
public:
  TTank();
  virtual ~TTank();

  virtual void ResetStatus(); // ]ZJA
  virtual TBullet* FireBullet(); // oglu

  __property int HP = {read = FHP, write = FHP};
  __property int MaxBulletNum = {read = FMaxBulletNum, write = FMaxBulletNum};
  __property bool SuperMode = {read = FSuperMode, write = FSuperMode};
};

// ڤZJO
class TMyTank : public TTank {
protected:
  // gIˬdĲok
  virtual bool CheckCollisions(int& x, int& y);

  // _Iˬd
  virtual bool CheckGemCollisions(int& x, int& y);
public:
  TMyTank();

  virtual void ResetStatus(); // ]ZJA
};

// ĤZJXͮɪ{{ĪGO
class TStar : public TSprite {
protected:
public:
  TStar();
};

// ĤZJO, |bXͮɥQ TStar Osy{{ĪG
class TETank : public TTank {
private:
  int FKindNo;
  TStar* FStar; // ҨϥΪ{{
protected:
public:
  TETank(int KindNo);

  virtual void Draw(TCanvas* Canvas);
  virtual void Move();

  virtual TBullet* FireBullet();
};

// zĪGO, HO
class TExplosion : public TSprite {
private:
  friend class TBullet;
  
  bool FLeadToOver; // ||ɭP Game Over
public:
  TExplosion();
};

// ɤΦaΪpz
class TSmallExplosion : public TExplosion {
protected:
public:
  TSmallExplosion();
};

// ZJjz
class TBigExplosion : public TExplosion {
protected:
public:
  TBigExplosion();
};

// luO, luѩZJ, FFۦR
class TBullet : public TSprite {
private:
  TTank* FTank; // ͦluZJ
  TExplosion* FExplosion; // luҲͪz
  bool FTankBulletNum_Dropped; // O_wNZJluƥػ

  // ҰzĪG (إz)
  void FireExplosion(const std::type_info& typeinfo);
  void LocateExplosion(TRect R); // Nzmxΰϰ줤
protected:
  // gIˬdĲok
  virtual bool CheckCollisions(int& x, int& y);

  // gҦIBzk, PɷsWPluIBzk
  virtual bool CheckBoundCollisions(int& x, int& y);
  virtual bool CheckCellCollisions(int Layer, int& x, int& y, TCellArray& Collisions);
  virtual bool CheckTankCollisions(int& x, int& y, TSpriteArray& Collisions);
  virtual bool CheckBulletCollisions(int& x, int& y, TSpriteArray& Collisions);
public:
  TBullet(TTank* ATank);
  virtual ~TBullet();

  virtual void Draw(TCanvas* Canvas);
  virtual void Move();

  __property TTank* Tank = {read = FTank, write = FTank};
};

// _O, |wɥX{beW
class TGem : public TSprite {
private:
  TGemKind FGemKind; // _
protected:
public:
  TGem(TGemKind AGemKind);
  virtual ~TGem();

  __property TGemKind GemKind = {read = FGemKind};
};
  
// ZJBz禡 (, øs, )
VAR void MoveTanks();
VAR void DrawTanks(TCanvas* Canvas, bool OnAir);
VAR void FreeTanks();

// luBz禡 (, øs, )
VAR void MoveBullets();
VAR void DrawBullets(TCanvas* Canvas); // øsҦlu
VAR void FreeBullets();
VAR void FreeBulletsForTank(TTank* ATank);

VAR TSpriteArray* Tanks(); // ҦZJ ( ]AڤZJ )
VAR TSpriteArray* Bullets(); // Ҧlu

VAR TGem* Gem INI(NULL);                     // _ (P@ɶu@)

#endif
